@@ -1,28 +1,25 @@
 /*
- * Copyright 2007 Peter Hutterer
+ * Copyright © 2007 Peter Hutterer
+ * Copyright © 2009 Red Hat, Inc.
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that
- * copyright notice and this permission notice appear in supporting
- * documentation.
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
  *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
- * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
- * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR
- * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
- * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
- * OTHER DEALINGS IN THE SOFTWARE.
- *
- * Except as contained in this notice, the name of the author shall
- * not be used in advertising or otherwise to promote the sale, use or
- * other dealings in this Software without prior written authorization
- * from the author.
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
  *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+ * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+ * DEALINGS IN THE SOFTWARE.
  */
 
 #include <ctype.h>
@@ -59,7 +56,7 @@ print_property(Display *dpy, XDevice* dev, Atom property)
     int                 act_format;
     unsigned long       nitems, bytes_after;
     unsigned char       *data, *ptr;
-    int                 j, done = False, size;
+    int                 j, done = False, size = 0;
 
     name = XGetAtomName(dpy, property);
     printf("\t%s (%ld):\t", name, property);
@@ -198,144 +195,6 @@ list_props_xi1(Display *dpy, int argc, char** argv, char* name, char *desc)
     return EXIT_SUCCESS;
 }
 
-int
-set_int_prop(Display *dpy, int argc, char** argv, char* n, char *desc)
-{
-    XDeviceInfo *info;
-    XDevice     *dev;
-    Atom         prop;
-    char        *name;
-    int          i;
-    char        *data;
-    int          format, nelements =  0;
-
-    if (argc < 3)
-    {
-        fprintf(stderr, "Usage: xinput %s %s\n", n, desc);
-        return EXIT_FAILURE;
-    }
-
-    info = find_device_info(dpy, argv[0], False);
-    if (!info)
-    {
-        fprintf(stderr, "unable to find device %s\n", argv[0]);
-        return EXIT_FAILURE;
-    }
-
-    dev = XOpenDevice(dpy, info->id);
-    if (!dev)
-    {
-        fprintf(stderr, "unable to open device %s\n", argv[0]);
-        return EXIT_FAILURE;
-    }
-
-    name = argv[1];
-
-    prop = parse_atom(dpy, name);
-
-    nelements = argc - 3;
-    format    = atoi(argv[2]);
-    if (format != 8 && format != 16 && format != 32)
-    {
-        fprintf(stderr, "Invalid format %d\n", format);
-        return EXIT_FAILURE;
-    }
-
-    data = calloc(nelements, sizeof(long));
-    for (i = 0; i < nelements; i++)
-    {
-        switch(format)
-        {
-            case 8:
-                *(((char*)data) + i) = atoi(argv[3 + i]);
-                break;
-            case 16:
-                *(((short*)data) + i) = atoi(argv[3 + i]);
-                break;
-            case 32:
-                *(((long*)data) + i) = atoi(argv[3 + i]);
-                break;
-        }
-    }
-
-    XChangeDeviceProperty(dpy, dev, prop, XA_INTEGER, format, PropModeReplace,
-                          (unsigned char*)data, nelements);
-
-    free(data);
-    XCloseDevice(dpy, dev);
-    return EXIT_SUCCESS;
-}
-
-int
-set_float_prop(Display *dpy, int argc, char** argv, char* n, char *desc)
-{
-    XDeviceInfo *info;
-    XDevice     *dev;
-    Atom         prop, float_atom;
-    char        *name;
-    int          i;
-    long        *data;
-    int          nelements =  0;
-    char*        endptr;
-
-    if (argc < 2)
-    {
-        fprintf(stderr, "Usage: xinput %s %s\n", n, desc);
-        return EXIT_FAILURE;
-    }
-
-    info = find_device_info(dpy, argv[0], False);
-    if (!info)
-    {
-        fprintf(stderr, "unable to find device %s\n", argv[0]);
-        return EXIT_FAILURE;
-    }
-
-    dev = XOpenDevice(dpy, info->id);
-    if (!dev)
-    {
-        fprintf(stderr, "unable to open device %s\n", argv[0]);
-        return EXIT_FAILURE;
-    }
-
-    name = argv[1];
-
-    prop = parse_atom(dpy, name);
-
-    nelements = argc - 2;
-
-    float_atom = XInternAtom(dpy, "FLOAT", False);
-
-    if (float_atom == (Atom)0)
-    {
-	fprintf(stderr, "no FLOAT atom present in server\n");
-	return EXIT_FAILURE;
-    }
-
-    if (sizeof(float) != 4)
-    {
-	fprintf(stderr, "sane FP required\n");
-	return EXIT_FAILURE;
-    }
-
-    data = calloc(nelements, sizeof(long));
-    for (i = 0; i < nelements; i++)
-    {
-        *((float*)(data + i)) = strtod(argv[2 + i], &endptr);
-	if(endptr == argv[2 + i]){
-	    fprintf(stderr, "argument %s could not be parsed\n", argv[2 + i]);
-	    return EXIT_FAILURE;
-	}
-    }
-
-    XChangeDeviceProperty(dpy, dev, prop, float_atom, 32, PropModeReplace,
-                          (unsigned char*)data, nelements);
-
-    free(data);
-    XCloseDevice(dpy, dev);
-    return EXIT_SUCCESS;
-}
-
 
 int watch_props(Display *dpy, int argc, char** argv, char* n, char *desc)
 {
@@ -415,83 +274,17 @@ delete_prop_xi1(Display *dpy, int argc, char** argv, char* n, char *desc)
     return EXIT_SUCCESS;
 }
 
-int
-set_atom_prop(Display *dpy, int argc, char** argv, char* n, char *desc)
-{
-    XDeviceInfo *info;
-    XDevice     *dev;
-    Atom         prop;
-    char        *name;
-    int          i, j;
-    Bool         is_atom;
-    Atom        *data;
-    int          nelements =  0;
-
-    if (argc < 3)
-    {
-        fprintf(stderr, "Usage: xinput %s %s\n", n, desc);
-        return EXIT_FAILURE;
-    }
-
-    info = find_device_info(dpy, argv[0], False);
-    if (!info)
-    {
-        fprintf(stderr, "unable to find device %s\n", argv[0]);
-        return EXIT_FAILURE;
-    }
-
-    dev = XOpenDevice(dpy, info->id);
-    if (!dev)
-    {
-        fprintf(stderr, "unable to open device %s\n", argv[0]);
-        return EXIT_FAILURE;
-    }
-
-    name = argv[1];
-
-    prop = parse_atom(dpy, name);
-
-    nelements = argc - 2;
-    data = calloc(nelements, sizeof(Atom));
-    for (i = 0; i < nelements; i++)
-    {
-        is_atom = True;
-        name = argv[2 + i];
-        for(j = 0; j < strlen(name); j++) {
-            if (!isdigit(name[j])) {
-                is_atom = False;
-                break;
-            }
-        }
-
-        if (!is_atom)
-            data[i] = XInternAtom(dpy, name, False);
-        else
-        {
-            data[i] = atoi(name);
-            XFree(XGetAtomName(dpy, data[i]));
-        }
-    }
-
-    XChangeDeviceProperty(dpy, dev, prop, XA_ATOM, 32, PropModeReplace,
-                          (unsigned char*)data, nelements);
-
-    free(data);
-    XCloseDevice(dpy, dev);
-    return EXIT_SUCCESS;
-}
-
 static int
-set_prop_xi1(Display *dpy, int argc, char **argv, char *n, char *desc)
+do_set_prop_xi1(Display *dpy, Atom type, int format, int argc, char **argv, char *n, char *desc)
 {
     XDeviceInfo  *info;
     XDevice      *dev;
     Atom          prop;
-    Atom          type;
+    Atom          old_type;
     char         *name;
     int           i;
     Atom          float_atom;
-    int           format, nelements = 0;
+    int           old_format, nelements = 0;
     unsigned long act_nitems, bytes_after;
     char         *endptr;
     union {
@@ -533,17 +326,26 @@ set_prop_xi1(Display *dpy, int argc, char **argv, char *n, char *desc)
     float_atom = XInternAtom(dpy, "FLOAT", False);
 
     nelements = argc - 2;
-    if (XGetDeviceProperty(dpy, dev, prop, 0, 0, False, AnyPropertyType,
-                           &type, &format, &act_nitems, &bytes_after, &data.c)
-            != Success) {
-        fprintf(stderr, "failed to get property type and format for %s\n", name);
-        return EXIT_FAILURE;
+    if (type == None || format == 0) {
+        if (XGetDeviceProperty(dpy, dev, prop, 0, 0, False, AnyPropertyType,
+                               &old_type, &old_format, &act_nitems,
+                               &bytes_after, &data.c) != Success) {
+            fprintf(stderr, "failed to get property type and format for %s\n",
+                    name);
+            return EXIT_FAILURE;
+        } else {
+            if (type == None)
+                type = old_type;
+            if (format == 0)
+                format = old_format;
+        }
+
+        XFree(data.c);
     }
 
-    XFree(data.c);
-
     if (type == None) {
-        fprintf(stderr, "property %s doesn't exist\n", name);
+        fprintf(stderr, "property %s doesn't exist, you need to specify "
+                "its type and format\n", name);
         return EXIT_FAILURE;
     }
 
@@ -656,10 +458,10 @@ print_property_xi2(Display *dpy, int deviceid, Atom property)
                     break;
                 case XA_ATOM:
                     {
-                        Atom a = *(Atom*)ptr;
-                        printf("\"%s\" (%d)",
+                        Atom a = *(uint32_t*)ptr;
+                        printf("\"%s\" (%ld)",
                                 (a) ? XGetAtomName(dpy, a) : "None",
-                                (int)a);
+                                a);
                         break;
                     }
                     break;
@@ -755,15 +557,15 @@ delete_prop_xi2(Display *dpy, int argc, char** argv, char* n, char *desc)
 }
 
 static int
-set_prop_xi2(Display *dpy, int argc, char **argv, char *n, char *desc)
+do_set_prop_xi2(Display *dpy, Atom type, int format, int argc, char **argv, char *n, char *desc)
 {
     XIDeviceInfo *info;
     Atom          prop;
-    Atom          type;
+    Atom          old_type;
     char         *name;
     int           i;
     Atom          float_atom;
-    int           format, nelements = 0;
+    int           old_format, nelements = 0;
     unsigned long act_nitems, bytes_after;
     char         *endptr;
     union {
@@ -797,17 +599,26 @@ set_prop_xi2(Display *dpy, int argc, char **argv, char *n, char *desc)
     float_atom = XInternAtom(dpy, "FLOAT", False);
 
     nelements = argc - 2;
-    if (XIGetProperty(dpy, info->deviceid, prop, 0, 0, False, AnyPropertyType,
-                      &type, &format, &act_nitems, &bytes_after, &data.c)
-            != Success) {
-        fprintf(stderr, "failed to get property type and format for %s\n", name);
-        return EXIT_FAILURE;
+    if (type == None || format == 0) {
+        if (XIGetProperty(dpy, info->deviceid, prop, 0, 0, False,
+                          AnyPropertyType, &old_type, &old_format, &act_nitems,
+                          &bytes_after, &data.c) != Success) {
+            fprintf(stderr, "failed to get property type and format for %s\n",
+                    name);
+            return EXIT_FAILURE;
+        } else {
+            if (type == None)
+                type = old_type;
+            if (format == 0)
+                format = old_format;
+        }
+
+        XFree(data.c);
     }
 
-    XFree(data.c);
-
     if (type == None) {
-        fprintf(stderr, "property %s doesn't exist\n", name);
+        fprintf(stderr, "property %s doesn't exist, you need to specify "
+                "its type and format\n", name);
         return EXIT_FAILURE;
     }
 
@@ -884,12 +695,104 @@ int delete_prop(Display *display, int argc, char *argv[], char *name,
 
 }
 
-int set_prop(Display *display, int argc, char *argv[], char *name,
-             char *desc)
+static int
+do_set_prop(Display *display, Atom type, int format, int argc, char *argv[], char *name, char *desc)
 {
 #ifdef HAVE_XI2
     if (xinput_version(display) == XI_2_Major)
-        return set_prop_xi2(display, argc, argv, name, desc);
+        return do_set_prop_xi2(display, type, format, argc, argv, name, desc);
 #endif
-    return set_prop_xi1(display, argc, argv, name, desc);
+    return do_set_prop_xi1(display, type, format, argc, argv, name, desc);
+}
+
+int
+set_atom_prop(Display *dpy, int argc, char** argv, char* n, char *desc)
+{
+    return do_set_prop(dpy, XA_ATOM, 32, argc, argv, n, desc);
+}
+
+int
+set_int_prop(Display *dpy, int argc, char** argv, char* n, char *desc)
+{
+    int          i;
+    int          format;
+
+    if (argc < 3)
+    {
+        fprintf(stderr, "Usage: xinput %s %s\n", n, desc);
+        return EXIT_FAILURE;
+    }
+
+    format    = atoi(argv[2]);
+    if (format != 8 && format != 16 && format != 32)
+    {
+        fprintf(stderr, "Invalid format %d\n", format);
+        return EXIT_FAILURE;
+    }
+
+    for (i = 3; i < argc; i++)
+        argv[i - 1] = argv[i];
+
+    return do_set_prop(dpy, XA_INTEGER, format, argc - 1, argv, n, desc);
+}
+
+int
+set_float_prop(Display *dpy, int argc, char** argv, char* n, char *desc)
+{
+    Atom float_atom = XInternAtom(dpy, "FLOAT", False);
+
+    if (sizeof(float) != 4)
+    {
+	fprintf(stderr, "sane FP required\n");
+	return EXIT_FAILURE;
+    }
+
+    return do_set_prop(dpy, float_atom, 32, argc, argv, n, desc);
+}
+
+int set_prop(Display *display, int argc, char *argv[], char *name,
+             char *desc)
+{
+    Atom type = None;
+    int format = 0;
+    int i = 0, j;
+
+    while (i < argc) {
+        char *option = strchr(argv[i], '=');
+        /* skip non-option arguments */
+        if (strncmp(argv[i], "--", 2) || !option) {
+            i++;
+            continue;
+        }
+
+        if (!strncmp(argv[i], "--type=", strlen("--type="))) {
+            if (!strcmp(option + 1, "int")) {
+                type = XA_INTEGER;
+            } else if (!strcmp(option + 1, "float")) {
+                type = XInternAtom(display, "FLOAT", False);
+                format = 32;
+            } else if (!strcmp(option + 1, "atom")) {
+                type = XA_ATOM;
+                format = 32;
+            } else {
+                fprintf(stderr, "unknown property type %s\n", option + 1);
+                return EXIT_FAILURE;
+            }
+        } else if (!strncmp(argv[i], "--format=", strlen("--format="))) {
+            format = atoi(option + 1);
+            if (format != 8 && format != 16 && format != 32) {
+                fprintf(stderr, "invalid property format %s\n", option + 1);
+                return EXIT_FAILURE;
+            }
+        } else {
+            fprintf(stderr, "invalid option %s\n", argv[i]);
+            return EXIT_FAILURE;
+        }
+
+        for (j = i; j + 1 < argc; j++)
+            argv[j] = argv[j + 1];
+        argc--;
+    }
+
+    return do_set_prop(display, type, format, argc, argv, name, desc);
 }
