@@ -68,7 +68,7 @@ static entry drivers[] =
      set_mode
     },
     {"list",
-     "[--short || --long] [<device name>...]",
+     "[--short || --long || --name-only || --id-only] [<device name>...]",
      list
     },
     {"query-state",
@@ -104,6 +104,10 @@ static entry drivers[] =
       "<device>",
       test_xi2,
     },
+    { "map-to-crtc",
+      "<device> <crtc name>",
+      map_to_crtc,
+    },
 #endif
     { "list-props",
       "<device> [<device> ...]",
@@ -139,8 +143,8 @@ static entry drivers[] =
 
 static const char version_id[] = VERSION;
 
-int
-print_version()
+static int
+print_version(void)
 {
     XExtensionVersion	*version;
     Display *display;
@@ -184,6 +188,23 @@ xinput_version(Display	*display)
 	XFree(version);
     }
 
+#if HAVE_XI2
+    /* Announce our supported version so the server treats us correctly. */
+    if (vers >= XI_2_Major)
+    {
+        int maj = 2,
+            min = 0;
+
+#if HAVE_XI21
+        min = 1;
+#elif HAVE_XI22
+        min = 2;
+#endif
+
+        XIQueryVersion(display, &maj, &min);
+    }
+#endif
+
     return vers;
 }
 
@@ -232,10 +253,42 @@ find_device_info(Display	*display,
 }
 
 #ifdef HAVE_XI2
+Bool is_pointer(int use)
+{
+    return use == XIMasterPointer || use == XISlavePointer;
+}
+
+Bool is_keyboard(int use)
+{
+    return use == XIMasterKeyboard || use == XISlaveKeyboard;
+}
+
+Bool device_matches(XIDeviceInfo *info, char *name)
+{
+    if (strcmp(info->name, name) == 0) {
+        return True;
+    }
+
+    if (strncmp(name, "pointer:", strlen("pointer:")) == 0 &&
+        strcmp(info->name, name + strlen("pointer:")) == 0 &&
+        is_pointer(info->use)) {
+        return True;
+    }
+
+    if (strncmp(name, "keyboard:", strlen("keyboard:")) == 0 &&
+        strcmp(info->name, name + strlen("keyboard:")) == 0 &&
+        is_keyboard(info->use)) {
+        return True;
+    }
+
+    return False;
+}
+
 XIDeviceInfo*
 xi2_find_device_info(Display *display, char *name)
 {
     XIDeviceInfo *info;
+    XIDeviceInfo *found = NULL;
     int ndevices;
     Bool is_id = True;
     int i, id = -1;
@@ -254,15 +307,22 @@ xi2_find_device_info(Display *display, char *name)
     info = XIQueryDevice(display, XIAllDevices, &ndevices);
     for(i = 0; i < ndevices; i++)
     {
-        if ((is_id && info[i].deviceid == id) ||
-                (!is_id && strcmp(info[i].name, name) == 0))
-        {
-            return &info[i];
+        if (is_id ? info[i].deviceid == id : device_matches (&info[i], name)) {
+            if (found) {
+                fprintf(stderr,
+                        "Warning: There are multiple devices matching '%s'.\n"
+                        "To ensure the correct one is selected, please use "
+                        "the device ID, or prefix the\ndevice name with "
+                        "'pointer:' or 'keyboard:' as appropriate.\n\n", name);
+                XIFreeDeviceInfo(info);
+                return NULL;
+            } else {
+                found = &info[i];
+            }
         }
     }
 
-    XIFreeDeviceInfo(info);
-    return NULL;
+    return found;
 }
 #endif
 
@@ -288,16 +348,20 @@ main(int argc, char * argv[])
     char        *func;
     int event, error;
 
-    if (argc < 2) {
-	usage();
-	return EXIT_FAILURE;
+    if (argc > 1) {
+	func = argv[1];
+	while(func[0] == '-') func++;
+    } else {
+	func = "list";
     }
 
-    func = argv[1];
-    while((*func) == '-') func++;
-
     if (strcmp("version", func) == 0) {
-        return print_version(argv[0]);
+        return print_version();
+    }
+
+    if (strcmp("help", func) == 0) {
+        usage();
+        return 0;
     }
 
     display = XOpenDisplay(NULL);
